from enum import Enum


class Vulnerability:
    def __init__(self, kind=None, description=None, transactions=None):
        self.type = kind
        self.description = description
        self.transactions = transactions
        self.tested = False
        self.confirmed = False

    def __str__(self):
        return """
        Type = {type}
        Description = {description}
        Transactions = {transactions}
        """.format(
            type=self.type, description=self.description, transactions=self.transactions
        )


class VulnerabilityType(Enum):
    KILL_ONLY = 0
    KILL_AND_WITHDRAW = 1
    ETHER_THEFT = 2



import argparse
import sys
import logging
from karl.output.stdout import Stdout
from karl.output.posturl import PostURL
from karl.output.folder import Folder
from karl.karl import Karl


def main():
    parser = argparse.ArgumentParser(
        description="Smart contract monitor using Mythril to find exploits"
    )

    # Ethereum node
    rpc = parser.add_argument_group("RPC options")
    rpc.add_argument(
        "--rpc",
        help="Custom RPC settings",
        metavar="HOST:PORT / ganache / infura-{mainnet, rinkeby, kovan, ropsten}",
    )
    rpc.add_argument(
        "--rpctls", type=bool, default=False, help="RPC connection over TLS"
    )
    rpc.add_argument(
        "--block",
        type=int,
        help="Start from this block, otherwise start from latest",
        metavar="NUMBER",
    )

    # Output
    output = parser.add_argument_group("Output")
    output.add_argument(
        "--output",
        help="Where to send results",
        default="stdout",
        metavar="Can be one of: stdout, posturl, folder",
    )
    output.add_argument(
        "--posturl",
        help="Send results to a RESTful url [when using `--output posturl`]",
    )
    output.add_argument(
        "--folder-output",
        help="Save files to this folder [when using `--output folder`]",
    )

    # Verbosity
    verbosity = parser.add_argument_group("Verbosity")
    verbosity.add_argument(
        "--verbose", "-v", action="count", help="Set verbosity level"
    )

    # Parse cli args
    args = parser.parse_args()

    # Set verbosity
    verbosity_default = logging.NOTSET
    verbosity_levels = {
        1: logging.CRITICAL,
        2: logging.ERROR,
        3: logging.WARNING,
        4: logging.INFO,
        5: logging.DEBUG,
    }
    verbose = (
        len(verbosity_levels)
        if args.verbose is not None and args.verbose > len(verbosity_levels)
        else args.verbose
    )

    output_destination = None
    if args.output == "stdout":
        output_destination = Stdout(
            verbosity=verbosity_levels.get(verbose, verbosity_default)
        )
    elif args.output == "posturl":
        if args.posturl is None:
            print(
                "No posturl specified. Set a destination with --posturl http://server:port/destination/url"
            )
            sys.exit()
        else:
            output_destination = PostURL(
                url=args.posturl,
                verbosity=verbosity_levels.get(verbose, verbosity_default),
            )
    elif args.output == "folder":
        output_destination = Folder(
            folder_path=args.folder_output,
            verbosity=verbosity_levels.get(verbose, verbosity_default),
        )

    if output_destination is None:
        print("Must pick an output destination with --output")
        sys.exit()

    # Start Karl
    try:
        karl = Karl(
            rpc=args.rpc,
            rpctls=args.rpctls,
            output=output_destination,
            verbosity=verbosity_levels.get(verbose, verbosity_default),
            block_number=args.block,
        )
        karl.run(forever=True)
    except Exception as e:
        print("[CLI] Exception:", e)


if __name__ == "__main__":
    main()



import time
import logging
import sys
from mythril.mythril import Mythril
from web3 import Web3
from karl.exceptions import RPCError
from karl.sandbox.sandbox import Sandbox
from karl.sandbox.exceptions import SandboxBaseException


logging.basicConfig(level=logging.INFO)


class Karl:
    """
    Karl main interface class.
    """

    def __init__(
        self,
        rpc=None,
        rpctls=False,
        block_number=None,
        output=None,
        verbosity=logging.INFO,
    ):
        """
            Initialize Karl with the received parameters
        """
        if rpc is None:
            raise (
                RPCError("Must provide a valid --rpc connection to an Ethereum node")
            )

        # Ethereum node to connect to
        self.rpc = rpc
        self.rpc_tls = rpctls
        # Send results to this output (could be stdout or restful url)
        self.output = output

        # ! hack to stop mythril logging
        logging.getLogger().setLevel(logging.CRITICAL)

        # Set logging verbosity
        self.logger = logging.getLogger("Karl")
        self.logger.setLevel(verbosity)

        # Init web3 client
        web3_rpc = None
        if rpc == "ganache":
            web3_rpc = "http://127.0.0.1:8545"
        else:
            infura_network = (
                rpc.split("infura-")[1] if rpc.startswith("infura-") else None
            )
            if infura_network in ["mainnet", "rinkeby", "kovan", "ropsten"]:
                web3_rpc = "https://{net}.infura.io".format(net=infura_network)
            else:
                try:
                    host, port = rpc.split(":")
                    if rpctls:
                        web3_rpc = "https://{host}:{port}".format(host=host, port=port)
                    else:
                        web3_rpc = "http://{host}:{port}".format(host=host, port=port)
                except ValueError:
                    raise RPCError(
                        "Invalid RPC argument provided '{}', use "
                        "'ganache', 'infura-[mainnet, rinkeby, kovan, ropsten]' "
                        "or HOST:PORT".format(rpc)
                    )
        if web3_rpc is None:
            raise RPCError(
                "Invalid RPC argument provided {}, use "
                "'ganache', 'infura-[mainnet, rinkeby, kovan, ropsten]' "
                "or HOST:PORT".format(rpc)
            )
        self.web3_rpc = web3_rpc
        self.web3 = Web3(Web3.HTTPProvider(web3_rpc, request_kwargs={"timeout": 60}))
        if self.web3 is None:
            raise RPCError(
                "Invalid RPC argument provided {}, use "
                "'ganache', 'infura-[mainnet, rinkeby, kovan, ropsten]' "
                "or HOST:PORT".format(rpc)
            )

        self.block_number = block_number or self.web3.eth.blockNumber

    def run(self, forever=True):
        self.logger.info("Starting scraping process")

        # TODO: Refactor try-except statements
        try:
            while forever:
                block = self.web3.eth.getBlock(
                    self.block_number, full_transactions=True
                )

                # If new block is not yet mined sleep and retry
                if block is None:
                    time.sleep(1)
                    continue

                self.logger.info(
                    "Processing block {block}".format(block=block.get("number"))
                )

                # Next block to scrape
                self.block_number += 1

                # For each transaction get the newly created accounts
                for t in block.get("transactions", []):
                    # If there is no to defined or to is reported as address(0x0)
                    # a new contract is created
                    if (t["to"] is not None) and (t["to"] != "0x0"):
                        continue
                    try:
                        receipt = self.web3.eth.getTransactionReceipt(t["hash"])
                        if (receipt is None) or (
                            receipt.get("contractAddress", None) is None
                        ):
                            self.logger.error(
                                "Receipt invalid for hash = {}".format(t["hash"].hex())
                            )
                            self.logger.error(receipt)
                            continue
                        address = str(receipt.get("contractAddress", None))
                        report = self._run_mythril(contract_address=address)

                        issues_num = len(report.issues)
                        if issues_num:
                            self.logger.info("Found %s issue(s)", issues_num)
                            self.output.send(report, contract_address=address)

                            self.logger.info("Firing up sandbox tester")
                            exploitable = self._run_sandbox(
                                block_number=block.get("number", None),
                                contract_address=address,
                                report=report,
                                rpc=self.web3_rpc,
                            )
                            if exploitable:
                                # TODO: Nice output
                                pass
                            else:
                                pass
                        else:
                            self.logger.info("No issues found")
                    except Exception as e:
                        self.logger.error("Exception: %s\n%s", e, sys.exc_info()[2])
        except Exception as e:
            self.logger.error("Exception: %s\n%s", e, sys.exc_info()[2])

    def _run_mythril(self, contract_address=None):
        myth = Mythril(onchain_storage_access=True, enable_online_lookup=True)
        myth.set_api_rpc(rpc=self.rpc, rpctls=self.rpc_tls)

        self.logger.info("Analyzing %s", contract_address)
        myth.load_from_address(contract_address)
        self.logger.debug("Running Mythril")

        return myth.fire_lasers(
            strategy="dfs",
            modules=["ether_thief", "suicide"],
            address=contract_address,
            execution_timeout=45,
            create_timeout=10,
            max_depth=22,
            transaction_count=3,
            verbose_report=True,
        )

    def _run_sandbox(
        self, block_number=None, contract_address=None, report=None, rpc=None
    ):
        exploitable = False
        try:
            sandbox = Sandbox(
                block_number=block_number,
                contract_address=contract_address,
                report=report,
                rpc=rpc,
                verbosity=self.logger.level,
            )
        except SandboxBaseException as e:
            self.logger.error(e)

        exploitable = sandbox.check_exploitability()

        return exploitable



# SandboxBaseException
class SandboxBaseException(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class BlockNumberInvalidException(SandboxBaseException):
    """Invalid specified block number"""

    def __init__(self):
        super().__init__("Invalid specified block number")


class ContractInvalidException(SandboxBaseException):
    """Invalid contract address"""

    def __init__(self):
        super().__init__("Invalid contract address")


class ReportInvalidException(SandboxBaseException):
    """Invalid report specified"""

    def __init__(self):
        super().__init__("Invalid report specified")


class RPCInvalidException(SandboxBaseException):
    """Invalid rpc endpoint specified"""

    def __init__(self):
        super().__init__("Invalid rpc endpoint specified")



from abc import ABC, abstractclassmethod
from mythril.analysis.report import Report


class OutputInterface(ABC):
    def __init__(self):
        super().__init__()

    @abstractclassmethod
    def send(self, report: Report, contract_address=""):
        pass



import logging
import sys
from urllib import request
from karl.output.output import OutputInterface
from mythril.analysis.report import Report
from karl.output.exceptions import PostURLInvalidURL


class PostURL(OutputInterface):
    def __init__(self, url=None, verbosity=logging.INFO):
        if url.startswith("file:/"):
            raise PostURLInvalidURL("Cannot post results to file:/ locations")
        self.url = url
        self.logger = logging.getLogger("PostURL")
        self.logger.setLevel(verbosity)
        self.logger.debug("PostURL enabled, sending to {url}".format(url=self.url))

    def send(self, report: Report, contract_address=""):
        try:
            req = request.Request(
                url=self.url, data=bytes(report.as_json(), "utf-8"), method="POST"
            )
            with request.urlopen(req) as f:
                self.logger.debug(f.read().decode("utf-8"))
        except Exception as e:
            self.logger.error("Exception: %s\n%s", e, sys.exc_info()[2])



import logging
import json
from karl.sandbox.exceptions import (
    BlockNumberInvalidException,
    ContractInvalidException,
    ReportInvalidException,
    RPCInvalidException,
)
from karl.sandbox.vulnerability import Vulnerability, VulnerabilityType
from karl.sandbox.ganache import Ganache
from web3 import Web3, HTTPProvider


class Sandbox:
    def __init__(
        self,
        block_number=None,
        contract_address=None,
        report=None,
        rpc=None,
        verbosity=logging.NOTSET,
    ):
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(verbosity)
        self.verbosity = verbosity

        if rpc is None:
            raise (RPCInvalidException)
        self.rpc = rpc

        if block_number is None:
            raise (BlockNumberInvalidException)
        self.block_number = block_number

        if contract_address is None:
            raise (ContractInvalidException)
        self.contract_address = contract_address

        if report is None:
            raise (ReportInvalidException)
        self.report = report

    def check_exploitability(self):
        exploitable = False

        hacker = Web3.toChecksumAddress(Ganache.accounts[9])
        feeder = Web3.toChecksumAddress(Ganache.accounts[8])

        # Parse report and generate vulnerability list
        vulns = []
        for i in range(0, len(self.report.issues)):
            issue = self.report.sorted_issues()[i]

            if "withdraw its balance" in issue["description"]:
                vuln_type = VulnerabilityType.KILL_AND_WITHDRAW
                description = (
                    "Looks line anyone can kill this contract and steal its balance."
                )
            elif "withdraw ETH" in issue["description"]:
                vuln_type = VulnerabilityType.ETHER_THEFT
                description = "Looks like anyone can withdraw ETH from this contract."
            else:
                vuln_type = VulnerabilityType.KILL_ONLY
                description = "Anybody can accidentally kill this contract."

            tx = issue["debug"].replace("\n", " ").replace("'", '"')
            transactions = json.loads(tx)

            # Build formatted transaction list
            transaction_list = []
            for _, t in transactions.items():
                tx = {
                    "from": hacker,
                    "to": self.contract_address,
                    # If a transaction has an argument that points
                    # to the hacker replace that
                    "data": t["calldata"].replace(
                        "deadbeefdeadbeefdeadbeefdeadbeefdeadbeef", hacker[2:]
                    ),
                    "value": int(t["call_value"], 16),
                }
                transaction_list.append(tx)

            vulns.append(
                Vulnerability(
                    kind=vuln_type,
                    description=description,
                    transactions=transaction_list,
                )
            )

        # Test each vulnerability
        for i, v in enumerate(vulns):
            self.logger.debug(v)

            # Set up new forked chain
            self.logger.debug("Forking chain at block {}".format(self.block_number))
            ganache = Ganache(
                internal_port=9545, rpc=self.rpc, verbosity=self.verbosity
            )
            w3 = Web3(HTTPProvider(ganache.internal_rpc))
            initial_balance = w3.eth.getBalance(hacker)

            # Force feed ether
            """
            pragma solidity 0.5.1;

            contract SelfDestructor {
                constructor(address payable _receiver) public payable {
                    selfdestruct(_receiver);
                }
            }
            """
            self.logger.debug("Feeding eth to contract")
            tx_feed_args = {
                "from": feeder,
                "data": "0x6080604052604051602080604c833981018060405260208110156021"
                "57600080fd5b81019080805190602001909291905050508073"
                "ffffffffffffffffffffffffffffffffffffffff16fffe"
                # Add the address of the contract that is force fed
                + str(24 * "0") + self.contract_address[2:].lower(),
                # Feed with 1 ETH
                "value": 10 ** 18,
            }
            self.logger.debug("Transaction = {}".format(tx_feed_args))
            tx_feed_hash = w3.eth.sendTransaction(tx_feed_args)
            tx_feed_receipt = w3.eth.waitForTransactionReceipt(tx_feed_hash, timeout=10)
            self.logger.debug("Fed ether in receipt = {}".format(tx_feed_receipt))

            # Sending transactions to chain
            for tx in v.transactions:
                self.logger.debug("Sending transaction")
                self.logger.debug("Transaction = {}".format(tx))
                tx_hash = w3.eth.sendTransaction(tx)
                tx_receipt = w3.eth.waitForTransactionReceipt(tx_hash, timeout=10)
                self.logger.debug("Receipt = {}".format(tx_receipt))

            final_balance = w3.eth.getBalance(hacker)
            if final_balance > initial_balance:
                print("POSSIBLE VULNERABILITY!")
                print(
                    "Initial balance = {}, final balance = {}".format(
                        initial_balance, final_balance
                    )
                )
                print(v)
                exploitable = True
            else:
                print("Doesn't have more ether after exploit")

            # Stop forked chain
            self.logger.debug("Stopping forked chain")
            ganache.stop()

        return exploitable



import logging
from karl.output.output import OutputInterface


class Folder(OutputInterface):
    def __init__(self, folder_path=None, verbosity=logging.INFO):
        if folder_path is None:
            folder_path = "./"

        self.folder_path = folder_path
        self.logger = logging.getLogger("Folder")
        self.logger.setLevel(verbosity)
        self.logger.debug(
            "Folder enabled, saving output to {path}".format(path=self.folder_path)
        )

    def send(self, report=dict(), contract_address=""):
        self.logger.debug(
            "Found {} issues for {}".format(len(report.issues), contract_address)
        )

        file_path = "{path}/{filename}.json".format(
            path=self.folder_path, filename=contract_address
        )
        with open(file_path, "w") as f:
            f.write(report.as_text())



import logging
from karl.output.output import OutputInterface


class Stdout(OutputInterface):
    def __init__(self, verbosity=logging.INFO):
        self.logger = logging.getLogger("Stdout")
        self.logger.setLevel(verbosity)
        self.logger.debug("Stdout enabled")

    def send(self, report=dict(), contract_address=""):
        self.logger.debug(
            "Found {} issues for {}".format(len(report.issues), contract_address)
        )
        self.logger.info(report.as_text())



import subprocess
import logging
import signal


class Ganache:
    accounts = [
        "0x90f8bf6a479f320ead074411a4b0e7944ea8c9c1",
        "0xffcf8fdee72ac11b5c542428b35eef5769c409f0",
        "0x22d491bde2303f2f43325b2108d26f1eaba1e32b",
        "0xe11ba2b4d45eaed5996cd0823791e0c93114882d",
        "0xd03ea8624c8c5987235048901fb614fdca89b117",
        "0x95ced938f7991cd0dfcb48f0a06a40fa1af46ebc",
        "0x3e5e9111ae8eb78fe1cc3bb8915d5d461f3ef9a9",
        "0x28a8746e75304c0780e011bed21c72cd78cd535e",
        "0xaca94ef8bd5ffee41947b4585a84bda5a3d3da6e",
        "0x1df62f291b2e969fb0849d99d9ce41e2f137006e",
    ]

    private_keys = [
        "0x4f3edf983ac636a65a842ce7c78d9aa706d3b113bce9c46f30d7d21715b23b1d",
        "0x6cbed15c793ce57650b9877cf6fa156fbef513c4e6134f022a85b1ffdd59b2a1",
        "0x6370fd033278c143179d81c5526140625662b8daa446c22ee2d73db3707e620c",
        "0x646f1ce2fdad0e6deeeb5c7e8e5543bdde65e86029e2fd9fc169899c440a7913",
        "0xadd53f9a7e588d003326d1cbf9e4a43c061aadd9bc938c843a79e7b4fd2ad743",
        "0x395df67f0c2d2d9fe1ad08d1bc8b6627011959b79c53d7dd6a3536a33ab8a4fd",
        "0xe485d098507f54e7733a205420dfddbe58db035fa577fc294ebd14db90767a52",
        "0xa453611d9419d0e56f499079478fd72c37b251a94bfde4d19872c44cf65386e3",
        "0x829e924fdf021ba3dbbc4225edfece9aca04b929d6e75613329ca6f1d31c0bb4",
        "0xb0057716d5917badaf911b193b12b910811c1497b5bada8d7711f758981c3773",
    ]

    def __init__(
        self,
        # Sandbox
        internal_host="localhost",
        internal_port=9545,
        deterministic=True,
        gas_price=1,
        # Blockchain to fork
        rpc=None,
        block_number=None,
        # Misc
        verbosity=logging.NOTSET,
    ):
        # Blockchain to fork
        self.rpc = rpc
        self.block_number = block_number

        # Where to start the sandbox
        self.internal_host = internal_host
        self.internal_port = internal_port
        self.internal_rpc = "http://{host}:{port}".format(
            host=self.internal_host, port=self.internal_port
        )

        # Gas price
        self.gas_price = gas_price

        # Account settings
        self.deterministic = deterministic

        # Verbosity
        self.verbosity = verbosity

        self.logger = logging.getLogger("Ganache")
        self.logger.setLevel(self.verbosity)

        # Building ganache-cli args
        self.logger.debug("Setting up ganache-cli args")
        args = ["ganache-cli"]

        # Host
        if self.internal_host is not None:
            args.extend(["-h", self.internal_host])

        # Port
        if self.internal_port is not None:
            args.extend(["-p", str(self.internal_port)])

        # Deterministic
        if self.deterministic:
            args.append("-d")

        # Gas price
        args.extend(["-g", str(self.gas_price)])

        # Fork blockchain
        if self.rpc is not None:
            if self.block_number is not None:
                args.extend(
                    [
                        "-f",
                        "{rpc}@{block_number}".format(
                            rpc=self.rpc, block_number=block_number
                        ),
                    ]
                )
            else:
                args.extend(["-f", "{rpc}".format(rpc=self.rpc)])

        self.logger.debug("Starting ganache with\n{}".format(" ".join(args)))
        self.process = subprocess.Popen(
            args, shell=False, universal_newlines=True, stdout=subprocess.PIPE
        )
        self.logger.debug(self.process.args)
        for l in self.process.stdout:
            self.logger.debug(l)
            # Wait for boot
            if "Listening on" in str(l):
                self.logger.debug(l)
                break

    def stop(self):
        self.process.send_signal(signal.SIGTERM)
